// Package parser parses a setup file's functions, types, and fields using an Abstract Syntax Tree.
package parser

import (
	"bytes"
	"fmt"
	"go/ast"
	"go/parser"
	"go/printer"
	"go/token"

	"github.com/switchupcb/copygen/cli/models"
	"github.com/switchupcb/copygen/cli/parser/options"
	"golang.org/x/tools/go/packages"
)

// Parser represents a parser that parses Abstract Syntax Tree data into models.
type Parser struct {
	Config  Config
	Options Options
	Pkgs    []*packages.Package
}

// Config represents a Parser's configuration.
type Config struct {
	// SetupFile represents the setup file as an Abstract Syntax Tree.
	SetupFile *ast.File

	// Fileset represents the parser's fileset.
	Fileset *token.FileSet
}

// Options represents a parser's options.
type Options struct {
	// commentOptionMap represents a map of comments (as text) to an option.
	CommentOptionMap map[string]*options.Option

	// convertOptions represents a global list of convert options (for convert functions).
	ConvertOptions []*options.Option
}

// GLOBAL VARIABLES.
var (
	// fieldcache represents a map of `go/types` Type strings to models.Field.
	//
	// fieldcache is used to prevent cyclic fields from incorrect assignment.
	//
	// fieldcache improves performance by parsing a unique type definition once per runtime.
	// definitions remain constant UNLESS the user modifies their modules during runtime.
	fieldcache map[string]*models.Field
)

// SetupCache sets up the parser's global cache.
func SetupCache() {
	if fieldcache == nil {
		fieldcache = make(map[string]*models.Field)
	}
}

// ResetCache resets the parser's global cache.
func ResetCache() {
	fieldcache = make(map[string]*models.Field)
}

// parserLoadMode represents the load mode required for sufficient information during package load.
const parserLoadMode = packages.NeedName + packages.NeedImports + packages.NeedDeps + packages.NeedTypes + packages.NeedSyntax + packages.NeedTypesInfo

// Parse parses a generator's setup file.
func Parse(gen *models.Generator) error {
	var err error
	p := new(Parser)
	p.Config.Fileset = token.NewFileSet()
	p.Config.SetupFile, err = parser.ParseFile(p.Config.Fileset, gen.Setpath, nil, parser.ParseComments)
	if err != nil {
		return fmt.Errorf("an error occurred parsing the specified .go setup file: %v\n%w", gen.Setpath, err)
	}

	// Parse the Keep (and set options in the process).
	if err := p.Keep(p.Config.SetupFile); err != nil {
		return fmt.Errorf("%w", err)
	}

	// Analyze the `type Copygen Interface` to create models.Function and models.Field objects.
	// load package types from the setup file (NOTE: loads different *ast.Files).
	cfg := &packages.Config{Mode: parserLoadMode}
	p.Pkgs, err = packages.Load(cfg, "file="+gen.Setpath)
	if err != nil {
		return fmt.Errorf("an error occurred while loading the packages for types.\n%w", err)
	}

	// find a new instance of a copygen AST since the old one has its comments removed.
	var newCopygen *ast.InterfaceType
	for _, decl := range p.Pkgs[0].Syntax[0].Decls {
		switch declaration := decl.(type) {
		case *ast.GenDecl:
			if it, ok := assertCopygenInterface(declaration); ok {
				newCopygen = it
				break
			}
		}
	}

	if newCopygen == nil {
		return fmt.Errorf("the \"type Copygen interface\" could not be found in the setup file")
	}

	// create the models.Function objects.
	SetupCache()
	if gen.Functions, err = p.parseFunctions(newCopygen); err != nil {
		return fmt.Errorf("%w", err)
	}

	// rename fields' packages using imports.
	p.setPackages(gen)

	// Write the Keep.
	buf := new(bytes.Buffer)
	buf.WriteString("// Code generated by github.com/switchupcb/copygen\n// DO NOT EDIT.\n\n")
	if err := printer.Fprint(buf, p.Config.Fileset, p.Config.SetupFile); err != nil {
		return fmt.Errorf("an error occurred writing the code that will be kept after generation\n%w", err)
	}
	gen.Keep = buf.Bytes()

	return nil
}

// setFieldPackage sets the packages for all fields in a generator using names from the setup file.
func (p *Parser) setPackages(gen *models.Generator) {

	// determine the output file package path.
	var outputPkgPath string
	pkgs, _ := packages.Load(&packages.Config{Mode: packages.NeedName}, "file="+gen.Outpath)
	if len(pkgs) > 0 {
		outputPkgPath = pkgs[0].PkgPath
	}
	setupPkgPath := p.Pkgs[0].PkgPath

	// adjust the packages of each field where necessary.
	imports := p.Config.SetupFile.Imports
	aliasImportMap := make(map[string]string, len(imports))
	for _, imp := range imports {
		if imp.Name != nil {
			aliasImportMap[imp.Path.Value[1:len(imp.Path.Value)-1]] = imp.Name.Name
		}
	}

	for _, function := range gen.Functions {
		functionTypes := [][]models.Type{
			function.From,
			function.To,
		}

		for _, types := range functionTypes {
			for _, t := range types {
				for _, field := range t.Field.AllFields(nil, nil) {

					// a generated file's package == setup file's package
					//
					// when the field is defined in the setup file (i.e `Collection`),
					// do NOT reference it by package in the generated file.
					if field.Import == setupPkgPath {
						field.Package = ""
						continue
					}

					// when a setup file imports the package it will output to,
					// do NOT reference the fields defined in the output package, by package.
					if outputPkgPath != "" && field.Import == outputPkgPath {
						field.Package = ""
						continue
					}

					// when a field's import uses an alias, reassign the package reference.
					if aliasPkg, ok := aliasImportMap[field.Import]; ok {
						field.Package = aliasPkg
						continue
					}
				}
			}
		}
	}
}
